@@ -5,6 +5,8 @@
 
 package org.postgresql.jdbc;
 
+import static org.postgresql.util.internal.Nullness.castNonNull;
+
 import org.postgresql.Driver;
 import org.postgresql.PGNotification;
 import org.postgresql.PGProperty;
@@ -15,7 +17,6 @@ import org.postgresql.core.CachedQuery;
 import org.postgresql.core.ConnectionFactory;
 import org.postgresql.core.Encoding;
 import org.postgresql.core.Oid;
-import org.postgresql.core.Provider;
 import org.postgresql.core.Query;
 import org.postgresql.core.QueryExecutor;
 import org.postgresql.core.ReplicationProtocol;
@@ -37,6 +38,13 @@ import org.postgresql.util.PGBinaryObject;
 import org.postgresql.util.PGobject;
 import org.postgresql.util.PSQLException;
 import org.postgresql.util.PSQLState;
+import org.postgresql.xml.DefaultPGXmlFactoryFactory;
+import org.postgresql.xml.LegacyInsecurePGXmlFactoryFactory;
+import org.postgresql.xml.PGXmlFactoryFactory;
+
+import org.checkerframework.checker.nullness.qual.Nullable;
+import org.checkerframework.checker.nullness.qual.PolyNull;
+import org.checkerframework.dataflow.qual.Pure;
 
 import java.io.IOException;
 import java.sql.Array;
@@ -58,6 +66,7 @@ import java.sql.Savepoint;
 import java.sql.Statement;
 import java.sql.Struct;
 import java.sql.Types;
+import java.util.Arrays;
 import java.util.Enumeration;
 import java.util.HashMap;
 import java.util.HashSet;
@@ -67,7 +76,6 @@ import java.util.NoSuchElementException;
 import java.util.Properties;
 import java.util.Set;
 import java.util.StringTokenizer;
-import java.util.TimeZone;
 import java.util.Timer;
 import java.util.TimerTask;
 import java.util.concurrent.Executor;
@@ -77,10 +85,16 @@ import java.util.logging.Logger;
 public class PgConnection implements BaseConnection {
 
   private static final Logger LOGGER = Logger.getLogger(PgConnection.class.getName());
-
+  private static final Set<Integer> SUPPORTED_BINARY_OIDS = getSupportedBinaryOids();
   private static final SQLPermission SQL_PERMISSION_ABORT = new SQLPermission("callAbort");
   private static final SQLPermission SQL_PERMISSION_NETWORK_TIMEOUT = new SQLPermission("setNetworkTimeout");
 
+  private enum ReadOnlyBehavior {
+    ignore,
+    transaction,
+    always;
+  }
+
   //
   // Data initialized on construction:
   //
@@ -89,7 +103,9 @@ public class PgConnection implements BaseConnection {
   /* URL we were created via */
   private final String creatingURL;
 
-  private Throwable openStackTrace;
+  private final ReadOnlyBehavior readOnlyBehavior;
+
+  private @Nullable Throwable openStackTrace;
 
   /* Actual network handler */
   private final QueryExecutor queryExecutor;
@@ -99,6 +115,10 @@ public class PgConnection implements BaseConnection {
   /* Query that runs ROLLBACK */
   private final Query rollbackQuery;
 
+  private final CachedQuery setSessionReadOnly;
+
+  private final CachedQuery setSessionNotReadOnly;
+
   private final TypeInfo typeCache;
 
   private boolean disableColumnSanitiser = false;
@@ -122,18 +142,19 @@ public class PgConnection implements BaseConnection {
   private boolean autoCommit = true;
   // Connection's readonly state.
   private boolean readOnly = false;
-
+  // Filter out database objects for which the current user has no privileges granted from the DatabaseMetaData
+  private boolean  hideUnprivilegedObjects ;
   // Bind String to UNSPECIFIED or VARCHAR?
   private final boolean bindStringAsVarchar;
 
   // Current warnings; there might be more on queryExecutor too.
-  private SQLWarning firstWarning = null;
+  private @Nullable SQLWarning firstWarning;
 
   // Timer for scheduling TimerTasks for this connection.
   // Only instantiated if a task is actually scheduled.
-  private volatile Timer cancelTimer = null;
+  private volatile @Nullable Timer cancelTimer;
 
-  private PreparedStatement checkConnectionQuery;
+  private @Nullable PreparedStatement checkConnectionQuery;
   /**
    * Replication protocol in current version postgresql(10devel) supports a limited number of
    * commands.
@@ -142,6 +163,9 @@ public class PgConnection implements BaseConnection {
 
   private final LruCache<FieldMetadata.Key, FieldMetadata> fieldMetadataCache;
 
+  private final @Nullable String xmlFactoryFactoryClass;
+  private @Nullable PGXmlFactoryFactory xmlFactoryFactory;
+
   final CachedQuery borrowQuery(String sql) throws SQLException {
     return queryExecutor.borrowQuery(sql);
   }
@@ -150,7 +174,8 @@ public class PgConnection implements BaseConnection {
     return queryExecutor.borrowCallableQuery(sql);
   }
 
-  private CachedQuery borrowReturningQuery(String sql, String[] columnNames) throws SQLException {
+  private CachedQuery borrowReturningQuery(String sql, String @Nullable [] columnNames)
+      throws SQLException {
     return queryExecutor.borrowReturningQuery(sql, columnNames);
   }
 
@@ -174,6 +199,7 @@ public class PgConnection implements BaseConnection {
   //
   // Ctor.
   //
+  @SuppressWarnings({"method.invocation.invalid", "argument.type.incompatible"})
   public PgConnection(HostSpec[] hostSpecs,
                       String user,
                       String database,
@@ -184,6 +210,8 @@ public class PgConnection implements BaseConnection {
 
     this.creatingURL = url;
 
+    this.readOnlyBehavior = getReadOnlyBehavior(PGProperty.READ_ONLY_MODE.get(info));
+
     setDefaultFetchSize(PGProperty.DEFAULT_ROW_FETCH_SIZE.getInt(info));
 
     setPrepareThreshold(PGProperty.PREPARE_THRESHOLD.getInt(info));
@@ -199,11 +227,16 @@ public class PgConnection implements BaseConnection {
       LOGGER.log(Level.WARNING, "Unsupported Server Version: {0}", queryExecutor.getServerVersion());
     }
 
+    setSessionReadOnly = createQuery("SET SESSION CHARACTERISTICS AS TRANSACTION READ ONLY", false, true);
+    setSessionNotReadOnly = createQuery("SET SESSION CHARACTERISTICS AS TRANSACTION READ WRITE", false, true);
+
     // Set read-only early if requested
     if (PGProperty.READ_ONLY.getBoolean(info)) {
       setReadOnly(true);
     }
 
+    this.hideUnprivilegedObjects = PGProperty.HIDE_UNPRIVILEGED_OBJECTS.getBoolean(info);
+
     Set<Integer> binaryOids = getBinaryOids(info);
 
     // split for receive and send for better control
@@ -246,12 +279,8 @@ public class PgConnection implements BaseConnection {
     }
 
     // Initialize timestamp stuff
-    timestampUtils = new TimestampUtils(!queryExecutor.getIntegerDateTimes(), new Provider<TimeZone>() {
-      @Override
-      public TimeZone get() {
-        return queryExecutor.getTimeZone();
-      }
-    });
+    timestampUtils = new TimestampUtils(!queryExecutor.getIntegerDateTimes(),
+        new QueryExecutorTimeZoneProvider(queryExecutor));
 
     // Initialize common queries.
     // isParameterized==true so full parse is performed and the engine knows the query
@@ -290,6 +319,47 @@ public class PgConnection implements BaseConnection {
         false);
 
     replicationConnection = PGProperty.REPLICATION.get(info) != null;
+
+    xmlFactoryFactoryClass = PGProperty.XML_FACTORY_FACTORY.get(info);
+  }
+
+  private static ReadOnlyBehavior getReadOnlyBehavior(String property) {
+    try {
+      return ReadOnlyBehavior.valueOf(property);
+    } catch (IllegalArgumentException e) {
+      try {
+        return ReadOnlyBehavior.valueOf(property.toLowerCase(Locale.US));
+      } catch (IllegalArgumentException e2) {
+        return ReadOnlyBehavior.transaction;
+      }
+    }
+  }
+
+  private static Set<Integer> getSupportedBinaryOids() {
+    return new HashSet<Integer>(Arrays.asList(
+        Oid.BYTEA,
+        Oid.INT2,
+        Oid.INT4,
+        Oid.INT8,
+        Oid.FLOAT4,
+        Oid.FLOAT8,
+        Oid.TIME,
+        Oid.DATE,
+        Oid.TIMETZ,
+        Oid.TIMESTAMP,
+        Oid.TIMESTAMPTZ,
+        Oid.BYTEA_ARRAY,
+        Oid.INT2_ARRAY,
+        Oid.INT4_ARRAY,
+        Oid.INT8_ARRAY,
+        Oid.OID_ARRAY,
+        Oid.FLOAT4_ARRAY,
+        Oid.FLOAT8_ARRAY,
+        Oid.VARCHAR_ARRAY,
+        Oid.TEXT_ARRAY,
+        Oid.POINT,
+        Oid.BOX,
+        Oid.UUID));
   }
 
   private static Set<Integer> getBinaryOids(Properties info) throws PSQLException {
@@ -297,31 +367,18 @@ public class PgConnection implements BaseConnection {
     // Formats that currently have binary protocol support
     Set<Integer> binaryOids = new HashSet<Integer>(32);
     if (binaryTransfer) {
-      binaryOids.add(Oid.BYTEA);
-      binaryOids.add(Oid.INT2);
-      binaryOids.add(Oid.INT4);
-      binaryOids.add(Oid.INT8);
-      binaryOids.add(Oid.FLOAT4);
-      binaryOids.add(Oid.FLOAT8);
-      binaryOids.add(Oid.TIME);
-      binaryOids.add(Oid.DATE);
-      binaryOids.add(Oid.TIMETZ);
-      binaryOids.add(Oid.TIMESTAMP);
-      binaryOids.add(Oid.TIMESTAMPTZ);
-      binaryOids.add(Oid.INT2_ARRAY);
-      binaryOids.add(Oid.INT4_ARRAY);
-      binaryOids.add(Oid.INT8_ARRAY);
-      binaryOids.add(Oid.FLOAT4_ARRAY);
-      binaryOids.add(Oid.FLOAT8_ARRAY);
-      binaryOids.add(Oid.VARCHAR_ARRAY);
-      binaryOids.add(Oid.TEXT_ARRAY);
-      binaryOids.add(Oid.POINT);
-      binaryOids.add(Oid.BOX);
-      binaryOids.add(Oid.UUID);
+      binaryOids.addAll(SUPPORTED_BINARY_OIDS);
+    }
+
+    String oids = PGProperty.BINARY_TRANSFER_ENABLE.get(info);
+    if (oids != null) {
+      binaryOids.addAll(getOidSet(oids));
+    }
+    oids = PGProperty.BINARY_TRANSFER_DISABLE.get(info);
+    if (oids != null) {
+      binaryOids.removeAll(getOidSet(oids));
     }
 
-    binaryOids.addAll(getOidSet(PGProperty.BINARY_TRANSFER_ENABLE.get(info)));
-    binaryOids.removeAll(getOidSet(PGProperty.BINARY_TRANSFER_DISABLE.get(info)));
     return binaryOids;
   }
 
@@ -358,7 +415,7 @@ public class PgConnection implements BaseConnection {
   /**
    * The current type mappings.
    */
-  protected Map<String, Class<?>> typemap;
+  protected Map<String, Class<?>> typemap = new HashMap<String, Class<?>>();
 
   @Override
   public Statement createStatement() throws SQLException {
@@ -431,7 +488,7 @@ public class PgConnection implements BaseConnection {
       addWarning(warnings);
     }
 
-    return stat.getResultSet();
+    return castNonNull(stat.getResultSet(), "hasResultSet==true, yet getResultSet()==null");
   }
 
   @Override
@@ -453,6 +510,24 @@ public class PgConnection implements BaseConnection {
     stmt.close();
   }
 
+  void execSQLUpdate(CachedQuery query) throws SQLException {
+    BaseStatement stmt = (BaseStatement) createStatement();
+    if (stmt.executeWithFlags(query, QueryExecutor.QUERY_NO_METADATA | QueryExecutor.QUERY_NO_RESULTS
+        | QueryExecutor.QUERY_SUPPRESS_BEGIN)) {
+      throw new PSQLException(GT.tr("A result was returned when none was expected."),
+          PSQLState.TOO_MANY_RESULTS);
+    }
+
+    // Transfer warnings to the connection, since the user never
+    // has a chance to see the statement itself.
+    SQLWarning warnings = stmt.getWarnings();
+    if (warnings != null) {
+      addWarning(warnings);
+    }
+
+    stmt.close();
+  }
+
   /**
    * <p>In SQL, a result table can be retrieved through a cursor that is named. The current row of a
    * result can be updated or deleted using a positioned update/delete statement that references the
@@ -474,7 +549,7 @@ public class PgConnection implements BaseConnection {
    * @return the current cursor name
    * @throws SQLException if a database access error occurs
    */
-  public String getCursorName() throws SQLException {
+  public @Nullable String getCursorName() throws SQLException {
     checkClosed();
     return null;
   }
@@ -511,7 +586,7 @@ public class PgConnection implements BaseConnection {
   }
 
   // This holds a reference to the Fastpath API if already open
-  private Fastpath fastpath = null;
+  private @Nullable Fastpath fastpath;
 
   public LargeObjectManager getLargeObjectAPI() throws SQLException {
     checkClosed();
@@ -522,7 +597,7 @@ public class PgConnection implements BaseConnection {
   }
 
   // This holds a reference to the LargeObject API if already open
-  private LargeObjectManager largeobject = null;
+  private @Nullable LargeObjectManager largeobject;
 
   /*
    * This method is used internally to return an object based around org.postgresql's more unique
@@ -539,7 +614,8 @@ public class PgConnection implements BaseConnection {
    * @exception SQLException if value is not correct for this type
    */
   @Override
-  public Object getObject(String type, String value, byte[] byteValue) throws SQLException {
+  public Object getObject(String type, @Nullable String value, byte @Nullable [] byteValue)
+      throws SQLException {
     if (typemap != null) {
       Class<?> c = typemap.get(type);
       if (c != null) {
@@ -631,9 +707,9 @@ public class PgConnection implements BaseConnection {
     Enumeration<?> e = info.propertyNames();
     while (e.hasMoreElements()) {
       String propertyName = (String) e.nextElement();
-      if (propertyName.startsWith("datatype.")) {
+      if (propertyName != null && propertyName.startsWith("datatype.")) {
         String typeName = propertyName.substring(9);
-        String className = info.getProperty(propertyName);
+        String className = castNonNull(info.getProperty(propertyName));
         Class<?> klass;
 
         try {
@@ -677,12 +753,12 @@ public class PgConnection implements BaseConnection {
   }
 
   @Override
-  public synchronized SQLWarning getWarnings() throws SQLException {
+  public synchronized @Nullable SQLWarning getWarnings() throws SQLException {
     checkClosed();
     SQLWarning newWarnings = queryExecutor.getWarnings(); // NB: also clears them.
     if (firstWarning == null) {
       firstWarning = newWarnings;
-    } else {
+    } else if (newWarnings != null) {
       firstWarning.setNextWarning(newWarnings); // Chain them on.
     }
 
@@ -692,6 +768,7 @@ public class PgConnection implements BaseConnection {
   @Override
   public synchronized void clearWarnings() throws SQLException {
     checkClosed();
+    //noinspection ThrowableNotThrown
     queryExecutor.getWarnings(); // Clear and discard.
     firstWarning = null;
   }
@@ -705,10 +782,8 @@ public class PgConnection implements BaseConnection {
           PSQLState.ACTIVE_SQL_TRANSACTION);
     }
 
-    if (readOnly != this.readOnly) {
-      String readOnlySql
-             = "SET SESSION CHARACTERISTICS AS TRANSACTION " + (readOnly ? "READ ONLY" : "READ WRITE");
-      execSQLUpdate(readOnlySql); // nb: no BEGIN triggered.
+    if (readOnly != this.readOnly && autoCommit && this.readOnlyBehavior == ReadOnlyBehavior.always) {
+      execSQLUpdate(readOnly ? setSessionReadOnly : setSessionNotReadOnly);
     }
 
     this.readOnly = readOnly;
@@ -721,6 +796,11 @@ public class PgConnection implements BaseConnection {
     return readOnly;
   }
 
+  @Override
+  public boolean hintReadOnly() {
+    return readOnly && readOnlyBehavior != ReadOnlyBehavior.ignore;
+  }
+
   @Override
   public void setAutoCommit(boolean autoCommit) throws SQLException {
     checkClosed();
@@ -733,6 +813,21 @@ public class PgConnection implements BaseConnection {
       commit();
     }
 
+    // if the connection is read only, we need to make sure session settings are
+    // correct when autocommit status changed
+    if (this.readOnly && readOnlyBehavior == ReadOnlyBehavior.always) {
+      // if we are turning on autocommit, we need to set session
+      // to read only
+      if (autoCommit) {
+        this.autoCommit = true;
+        execSQLUpdate(setSessionReadOnly);
+      } else {
+        // if we are turning auto commit off, we need to
+        // disable session
+        execSQLUpdate(setSessionNotReadOnly);
+      }
+    }
+
     this.autoCommit = autoCommit;
     LOGGER.log(Level.FINE, "  setAutoCommit = {0}", autoCommit);
   }
@@ -858,7 +953,7 @@ public class PgConnection implements BaseConnection {
     LOGGER.log(Level.FINE, "  setTransactionIsolation = {0}", isolationLevelName);
   }
 
-  protected String getIsolationLevelName(int level) {
+  protected @Nullable String getIsolationLevelName(int level) {
     switch (level) {
       case Connection.TRANSACTION_READ_COMMITTED:
         return "READ COMMITTED";
@@ -883,6 +978,10 @@ public class PgConnection implements BaseConnection {
     return queryExecutor.getDatabase();
   }
 
+  public boolean getHideUnprivilegedObjects() {
+    return hideUnprivilegedObjects;
+  }
+
   /**
    * <p>Overrides finalize(). If called, it closes the connection.</p>
    *
@@ -950,13 +1049,14 @@ public class PgConnection implements BaseConnection {
     return haveMinimumServerVersion(ver.getVersionNum());
   }
 
+  @Pure
   @Override
   public Encoding getEncoding() {
     return queryExecutor.getEncoding();
   }
 
   @Override
-  public byte[] encodeString(String str) throws SQLException {
+  public byte @PolyNull [] encodeString(@PolyNull String str) throws SQLException {
     try {
       return getEncoding().encode(str);
     } catch (IOException ioe) {
@@ -977,7 +1077,7 @@ public class PgConnection implements BaseConnection {
   }
 
   // This is a cache of the DatabaseMetaData instance for this connection
-  protected java.sql.DatabaseMetaData metadata;
+  protected java.sql.@Nullable DatabaseMetaData metadata;
 
   @Override
   public boolean isClosed() throws SQLException {
@@ -1001,7 +1101,7 @@ public class PgConnection implements BaseConnection {
     getQueryExecutor().processNotifies(timeoutMillis);
     // Backwards-compatibility hand-holding.
     PGNotification[] notifications = queryExecutor.getNotifications();
-    return (notifications.length == 0 ? null : notifications);
+    return notifications;
   }
 
   /**
@@ -1065,7 +1165,7 @@ public class PgConnection implements BaseConnection {
     return bindStringAsVarchar;
   }
 
-  private CopyManager copyManager = null;
+  private @Nullable CopyManager copyManager;
 
   public CopyManager getCopyAPI() throws SQLException {
     checkClosed();
@@ -1161,33 +1261,6 @@ public class PgConnection implements BaseConnection {
     return new PGReplicationConnectionImpl(this);
   }
 
-  private static void appendArray(StringBuilder sb, Object elements, char delim) {
-    sb.append('{');
-
-    int nElements = java.lang.reflect.Array.getLength(elements);
-    for (int i = 0; i < nElements; i++) {
-      if (i > 0) {
-        sb.append(delim);
-      }
-
-      Object o = java.lang.reflect.Array.get(elements, i);
-      if (o == null) {
-        sb.append("NULL");
-      } else if (o.getClass().isArray()) {
-        final PrimitiveArraySupport arraySupport = PrimitiveArraySupport.getArraySupport(o);
-        if (arraySupport != null) {
-          arraySupport.appendArray(sb, delim, o);
-        } else {
-          appendArray(sb, o, delim);
-        }
-      } else {
-        String s = o.toString();
-        PgArray.escapeArrayElement(sb, s);
-      }
-    }
-    sb.append('}');
-  }
-
   // Parse a "dirty" integer surrounded by non-numeric characters
   private static int integerPart(String dirtyString) {
     int start = 0;
@@ -1246,7 +1319,7 @@ public class PgConnection implements BaseConnection {
     LOGGER.log(Level.FINE, "  setTypeMap = {0}", map);
   }
 
-  protected Array makeArray(int oid, String fieldString) throws SQLException {
+  protected Array makeArray(int oid, @Nullable String fieldString) throws SQLException {
     return new PgArray(this, oid, fieldString);
   }
 
@@ -1292,8 +1365,9 @@ public class PgConnection implements BaseConnection {
     throw org.postgresql.Driver.notImplemented(this.getClass(), "createStruct(String, Object[])");
   }
 
+  @SuppressWarnings({ "rawtypes", "unchecked" })
   @Override
-  public Array createArrayOf(String typeName, Object elements) throws SQLException {
+  public Array createArrayOf(String typeName, @Nullable Object elements) throws SQLException {
     checkClosed();
 
     final TypeInfo typeInfo = getTypeInfo();
@@ -1310,52 +1384,19 @@ public class PgConnection implements BaseConnection {
       return makeArray(oid, null);
     }
 
-    final String arrayString;
-
-    final PrimitiveArraySupport arraySupport = PrimitiveArraySupport.getArraySupport(elements);
-
-    if (arraySupport != null) {
-      // if the oid for the given type matches the default type, we might be
-      // able to go straight to binary representation
-      if (oid == arraySupport.getDefaultArrayTypeOid(typeInfo) && arraySupport.supportBinaryRepresentation()
-          && getPreferQueryMode() != PreferQueryMode.SIMPLE) {
-        return new PgArray(this, oid, arraySupport.toBinaryRepresentation(this, elements));
-      }
-      arrayString = arraySupport.toArrayString(delim, elements);
-    } else {
-      final Class<?> clazz = elements.getClass();
-      if (!clazz.isArray()) {
-        throw new PSQLException(GT.tr("Invalid elements {0}", elements), PSQLState.INVALID_PARAMETER_TYPE);
-      }
-      StringBuilder sb = new StringBuilder();
-      appendArray(sb, elements, delim);
-      arrayString = sb.toString();
+    final ArrayEncoding.ArrayEncoder arraySupport = ArrayEncoding.getArrayEncoder(elements);
+    if (arraySupport.supportBinaryRepresentation(oid) && getPreferQueryMode() != PreferQueryMode.SIMPLE) {
+      return new PgArray(this, oid, arraySupport.toBinaryRepresentation(this, elements, oid));
     }
 
+    final String arrayString = arraySupport.toArrayString(delim, elements);
     return makeArray(oid, arrayString);
   }
 
   @Override
-  public Array createArrayOf(String typeName, Object[] elements) throws SQLException {
-    checkClosed();
-
-    int oid = getTypeInfo().getPGArrayType(typeName);
-
-    if (oid == Oid.UNSPECIFIED) {
-      throw new PSQLException(
-          GT.tr("Unable to find server array type for provided name {0}.", typeName),
-          PSQLState.INVALID_NAME);
-    }
-
-    if (elements == null) {
-      return makeArray(oid, null);
-    }
-
-    char delim = getTypeInfo().getArrayDelimiter(oid);
-    StringBuilder sb = new StringBuilder();
-    appendArray(sb, elements, delim);
-
-    return makeArray(oid, sb.toString());
+  public Array createArrayOf(String typeName, @Nullable Object @Nullable [] elements)
+      throws SQLException {
+    return createArrayOf(typeName, (Object) elements);
   }
 
   @Override
@@ -1367,10 +1408,17 @@ public class PgConnection implements BaseConnection {
     if (isClosed()) {
       return false;
     }
+    boolean changedNetworkTimeout = false;
     try {
-      int savedNetworkTimeOut = getNetworkTimeout();
+      int oldNetworkTimeout = getNetworkTimeout();
+      int newNetworkTimeout = (int) Math.min(timeout * 1000L, Integer.MAX_VALUE);
       try {
-        setNetworkTimeout(null, timeout * 1000);
+        // change network timeout only if the new value is less than the current
+        // (zero means infinite timeout)
+        if (newNetworkTimeout != 0 && (oldNetworkTimeout == 0 || newNetworkTimeout < oldNetworkTimeout)) {
+          changedNetworkTimeout = true;
+          setNetworkTimeout(null, newNetworkTimeout);
+        }
         if (replicationConnection) {
           Statement statement = createStatement();
           statement.execute("IDENTIFY_SYSTEM");
@@ -1384,7 +1432,9 @@ public class PgConnection implements BaseConnection {
         }
         return true;
       } finally {
-        setNetworkTimeout(null, savedNetworkTimeOut);
+        if (changedNetworkTimeout) {
+          setNetworkTimeout(null, oldNetworkTimeout);
+        }
       }
     } catch (SQLException e) {
       if (PSQLState.IN_FAILED_SQL_TRANSACTION.getState().equals(e.getSQLState())) {
@@ -1397,7 +1447,7 @@ public class PgConnection implements BaseConnection {
   }
 
   @Override
-  public void setClientInfo(String name, String value) throws SQLClientInfoException {
+  public void setClientInfo(String name, @Nullable String value) throws SQLClientInfoException {
     try {
       checkClosed();
     } catch (final SQLException cause) {
@@ -1466,7 +1516,7 @@ public class PgConnection implements BaseConnection {
   }
 
   @Override
-  public String getClientInfo(String name) throws SQLException {
+  public @Nullable String getClientInfo(String name) throws SQLException {
     checkClosed();
     clientInfo.put("ApplicationName", queryExecutor.getApplicationName());
     return clientInfo.getProperty(name);
@@ -1499,7 +1549,7 @@ public class PgConnection implements BaseConnection {
     throw new SQLException("Cannot unwrap to " + iface.getName());
   }
 
-  public String getSchema() throws SQLException {
+  public @Nullable String getSchema() throws SQLException {
     checkClosed();
     Statement stmt = createStatement();
     try {
@@ -1517,7 +1567,7 @@ public class PgConnection implements BaseConnection {
     }
   }
 
-  public void setSchema(String schema) throws SQLException {
+  public void setSchema(@Nullable String schema) throws SQLException {
     checkClosed();
     Statement stmt = createStatement();
     try {
@@ -1556,7 +1606,8 @@ public class PgConnection implements BaseConnection {
     executor.execute(command);
   }
 
-  public void setNetworkTimeout(Executor executor /*not used*/, int milliseconds) throws SQLException {
+  public void setNetworkTimeout(@Nullable Executor executor /*not used*/, int milliseconds)
+      throws SQLException {
     checkClosed();
 
     if (milliseconds < 0) {
@@ -1702,7 +1753,7 @@ public class PgConnection implements BaseConnection {
   }
 
   @Override
-  public PreparedStatement prepareStatement(String sql, int[] columnIndexes) throws SQLException {
+  public PreparedStatement prepareStatement(String sql, int @Nullable [] columnIndexes) throws SQLException {
     if (columnIndexes != null && columnIndexes.length == 0) {
       return prepareStatement(sql);
     }
@@ -1713,7 +1764,7 @@ public class PgConnection implements BaseConnection {
   }
 
   @Override
-  public PreparedStatement prepareStatement(String sql, String[] columnNames) throws SQLException {
+  public PreparedStatement prepareStatement(String sql, String @Nullable[] columnNames) throws SQLException {
     if (columnNames != null && columnNames.length == 0) {
       return prepareStatement(sql);
     }
@@ -1740,8 +1791,43 @@ public class PgConnection implements BaseConnection {
   }
 
   @Override
-  public final String getParameterStatus(String parameterName) {
+  public final @Nullable String getParameterStatus(String parameterName) {
     return queryExecutor.getParameterStatus(parameterName);
   }
 
+  @Override
+  public PGXmlFactoryFactory getXmlFactoryFactory() throws SQLException {
+    PGXmlFactoryFactory xmlFactoryFactory = this.xmlFactoryFactory;
+    if (xmlFactoryFactory != null) {
+      return xmlFactoryFactory;
+    }
+    if (xmlFactoryFactoryClass == null || xmlFactoryFactoryClass.equals("")) {
+      xmlFactoryFactory = DefaultPGXmlFactoryFactory.INSTANCE;
+    } else if (xmlFactoryFactoryClass.equals("LEGACY_INSECURE")) {
+      xmlFactoryFactory = LegacyInsecurePGXmlFactoryFactory.INSTANCE;
+    } else {
+      Class<?> clazz;
+      try {
+        clazz = Class.forName(xmlFactoryFactoryClass);
+      } catch (ClassNotFoundException ex) {
+        throw new PSQLException(
+            GT.tr("Could not instantiate xmlFactoryFactory: {0}", xmlFactoryFactoryClass),
+            PSQLState.INVALID_PARAMETER_VALUE, ex);
+      }
+      if (!clazz.isAssignableFrom(PGXmlFactoryFactory.class)) {
+        throw new PSQLException(
+            GT.tr("Connection property xmlFactoryFactory must implement PGXmlFactoryFactory: {0}", xmlFactoryFactoryClass),
+            PSQLState.INVALID_PARAMETER_VALUE);
+      }
+      try {
+        xmlFactoryFactory = (PGXmlFactoryFactory) clazz.newInstance();
+      } catch (Exception ex) {
+        throw new PSQLException(
+            GT.tr("Could not instantiate xmlFactoryFactory: {0}", xmlFactoryFactoryClass),
+            PSQLState.INVALID_PARAMETER_VALUE, ex);
+      }
+    }
+    this.xmlFactoryFactory = xmlFactoryFactory;
+    return xmlFactoryFactory;
+  }
 }
