@@ -5,6 +5,8 @@
 
 package org.postgresql.jdbc;
 
+import static org.postgresql.util.internal.Nullness.castNonNull;
+
 import org.postgresql.Driver;
 import org.postgresql.core.BaseConnection;
 import org.postgresql.core.BaseStatement;
@@ -16,10 +18,16 @@ import org.postgresql.core.QueryExecutor;
 import org.postgresql.core.ResultCursor;
 import org.postgresql.core.ResultHandlerBase;
 import org.postgresql.core.SqlCommand;
+import org.postgresql.core.Tuple;
 import org.postgresql.util.GT;
 import org.postgresql.util.PSQLException;
 import org.postgresql.util.PSQLState;
 
+import org.checkerframework.checker.index.qual.NonNegative;
+import org.checkerframework.checker.lock.qual.GuardedBy;
+import org.checkerframework.checker.nullness.qual.Nullable;
+import org.checkerframework.checker.nullness.qual.RequiresNonNull;
+
 import java.sql.Connection;
 import java.sql.ResultSet;
 import java.sql.SQLException;
@@ -42,8 +50,8 @@ public class PgStatement implements Statement, BaseStatement {
   // only for testing purposes. even single shot statements will use binary transfers
   private boolean forceBinaryTransfers = DEFAULT_FORCE_BINARY_TRANSFERS;
 
-  protected ArrayList<Query> batchStatements = null;
-  protected ArrayList<ParameterList> batchParameters = null;
+  protected @Nullable ArrayList<Query> batchStatements = null;
+  protected @Nullable ArrayList<@Nullable ParameterList> batchParameters = null;
   protected final int resultsettype; // the resultset type to return (ResultSet.TYPE_xxx)
   protected final int concurrency; // is it updateable or not? (ResultSet.CONCUR_xxx)
   private final int rsHoldability;
@@ -59,9 +67,10 @@ public class PgStatement implements Statement, BaseStatement {
    * cancelTask was created. Note: the field must be set/get/compareAndSet via
    * {@link #CANCEL_TIMER_UPDATER} as per {@link AtomicReferenceFieldUpdater} javadoc.
    */
-  private volatile TimerTask cancelTimerTask = null;
-  private static final AtomicReferenceFieldUpdater<PgStatement, TimerTask> CANCEL_TIMER_UPDATER =
-      AtomicReferenceFieldUpdater.newUpdater(PgStatement.class, TimerTask.class, "cancelTimerTask");
+  private volatile @Nullable TimerTask cancelTimerTask = null;
+  private static final AtomicReferenceFieldUpdater<PgStatement, @Nullable TimerTask> CANCEL_TIMER_UPDATER =
+      AtomicReferenceFieldUpdater.<PgStatement, @Nullable TimerTask>newUpdater(
+          PgStatement.class, TimerTask.class, "cancelTimerTask");
 
   /**
    * Protects statement from out-of-order cancels. It protects from both
@@ -98,7 +107,7 @@ public class PgStatement implements Statement, BaseStatement {
   /**
    * The warnings chain.
    */
-  protected volatile PSQLWarningWrapper warnings = null;
+  protected volatile @Nullable PSQLWarningWrapper warnings = null;
 
   /**
    * Maximum number of rows to return, 0 = unlimited.
@@ -120,22 +129,23 @@ public class PgStatement implements Statement, BaseStatement {
   /**
    * The current results.
    */
-  protected ResultWrapper result = null;
+  protected @Nullable ResultWrapper result = null;
 
   /**
    * The first unclosed result.
    */
-  protected ResultWrapper firstUnclosedResult = null;
+  protected @Nullable @GuardedBy("<self>") ResultWrapper firstUnclosedResult = null;
 
   /**
    * Results returned by a statement that wants generated keys.
    */
-  protected ResultWrapper generatedKeys = null;
+  protected @Nullable ResultWrapper generatedKeys = null;
 
   protected int mPrepareThreshold; // Reuse threshold to enable use of PREPARE
 
   protected int maxFieldSize = 0;
 
+  @SuppressWarnings("method.invocation.invalid")
   PgStatement(PgConnection c, int rsType, int rsConcurrency, int rsHoldability)
       throws SQLException {
     this.connection = c;
@@ -147,8 +157,9 @@ public class PgStatement implements Statement, BaseStatement {
     this.rsHoldability = rsHoldability;
   }
 
-  public ResultSet createResultSet(Query originalQuery, Field[] fields, List<byte[][]> tuples,
-      ResultCursor cursor) throws SQLException {
+  @SuppressWarnings("method.invocation.invalid")
+  public ResultSet createResultSet(@Nullable Query originalQuery, Field[] fields, List<Tuple> tuples,
+      @Nullable ResultCursor cursor) throws SQLException {
     PgResultSet newResult = new PgResultSet(originalQuery, this, fields, tuples, cursor,
         getMaxRows(), getMaxFieldSize(), getResultSetType(), getResultSetConcurrency(),
         getResultSetHoldability());
@@ -161,11 +172,11 @@ public class PgStatement implements Statement, BaseStatement {
     return connection;
   }
 
-  public String getFetchingCursorName() {
+  public @Nullable String getFetchingCursorName() {
     return null;
   }
 
-  public int getFetchSize() {
+  public @NonNegative int getFetchSize() {
     return fetchSize;
   }
 
@@ -182,10 +193,10 @@ public class PgStatement implements Statement, BaseStatement {
    * ResultHandler implementations for updates, queries, and either-or.
    */
   public class StatementResultHandler extends ResultHandlerBase {
-    private ResultWrapper results;
-    private ResultWrapper lastResult;
+    private @Nullable ResultWrapper results;
+    private @Nullable ResultWrapper lastResult;
 
-    ResultWrapper getResults() {
+    @Nullable ResultWrapper getResults() {
       return results;
     }
 
@@ -193,13 +204,13 @@ public class PgStatement implements Statement, BaseStatement {
       if (results == null) {
         lastResult = results = newResult;
       } else {
-        lastResult.append(newResult);
+        castNonNull(lastResult).append(newResult);
       }
     }
 
     @Override
-    public void handleResultRows(Query fromQuery, Field[] fields, List<byte[][]> tuples,
-        ResultCursor cursor) {
+    public void handleResultRows(Query fromQuery, Field[] fields, List<Tuple> tuples,
+        @Nullable ResultCursor cursor) {
       try {
         ResultSet rs = PgStatement.this.createResultSet(fromQuery, fields, tuples, cursor);
         append(new ResultWrapper(rs));
@@ -209,7 +220,7 @@ public class PgStatement implements Statement, BaseStatement {
     }
 
     @Override
-    public void handleCommandStatus(String status, int updateCount, long insertOID) {
+    public void handleCommandStatus(String status, long updateCount, long insertOID) {
       append(new ResultWrapper(updateCount, insertOID));
     }
 
@@ -232,22 +243,24 @@ public class PgStatement implements Statement, BaseStatement {
   protected ResultSet getSingleResultSet() throws SQLException {
     synchronized (this) {
       checkClosed();
+      ResultWrapper result = castNonNull(this.result);
       if (result.getNext() != null) {
         throw new PSQLException(GT.tr("Multiple ResultSets were returned by the query."),
             PSQLState.TOO_MANY_RESULTS);
       }
 
-      return result.getResultSet();
+      return castNonNull(result.getResultSet(), "result.getResultSet()");
     }
   }
 
   @Override
   public int executeUpdate(String sql) throws SQLException {
     executeWithFlags(sql, QueryExecutor.QUERY_NO_RESULTS);
-    return getNoResultUpdateCount();
+    checkNoResultUpdate();
+    return getUpdateCount();
   }
 
-  protected int getNoResultUpdateCount() throws SQLException {
+  protected final void checkNoResultUpdate() throws SQLException {
     synchronized (this) {
       checkClosed();
       ResultWrapper iter = result;
@@ -255,12 +268,9 @@ public class PgStatement implements Statement, BaseStatement {
         if (iter.getResultSet() != null) {
           throw new PSQLException(GT.tr("A result was returned when none was expected."),
               PSQLState.TOO_MANY_RESULTS);
-
         }
         iter = iter.getNext();
       }
-
-      return getUpdateCount();
     }
   }
 
@@ -274,7 +284,8 @@ public class PgStatement implements Statement, BaseStatement {
     return executeCachedSql(sql, flags, NO_RETURNING_COLUMNS);
   }
 
-  private boolean executeCachedSql(String sql, int flags, String[] columnNames) throws SQLException {
+  private boolean executeCachedSql(String sql, int flags,
+      String @Nullable [] columnNames) throws SQLException {
     PreferQueryMode preferQueryMode = connection.getPreferQueryMode();
     // Simple statements should not replace ?, ? with $1, $2
     boolean shouldUseParameterized = false;
@@ -321,6 +332,25 @@ public class PgStatement implements Statement, BaseStatement {
         PSQLState.WRONG_OBJECT_TYPE);
   }
 
+  /*
+  If there are multiple result sets we close any that have been processed and left open
+  by the client.
+   */
+  private void closeUnclosedProcessedResults() throws SQLException {
+    synchronized (this) {
+      ResultWrapper resultWrapper = this.firstUnclosedResult;
+      ResultWrapper currentResult = this.result;
+      for (; resultWrapper != currentResult && resultWrapper != null;
+           resultWrapper = resultWrapper.getNext()) {
+        PgResultSet rs = (PgResultSet) resultWrapper.getResultSet();
+        if (rs != null) {
+          rs.closeInternally();
+        }
+      }
+      firstUnclosedResult = resultWrapper;
+    }
+  }
+
   protected void closeForNextExecution() throws SQLException {
 
     // Every statement execution clears any previous warnings.
@@ -328,20 +358,20 @@ public class PgStatement implements Statement, BaseStatement {
 
     // Close any existing resultsets associated with this statement.
     synchronized (this) {
-      while (firstUnclosedResult != null) {
-        ResultSet rs = firstUnclosedResult.getResultSet();
-        if (rs != null) {
-          rs.close();
-        }
-        firstUnclosedResult = firstUnclosedResult.getNext();
+      closeUnclosedProcessedResults();
+
+      if ( this.result != null && this.result.getResultSet() != null ) {
+        this.result.getResultSet().close();
       }
       result = null;
 
+      ResultWrapper generatedKeys = this.generatedKeys;
       if (generatedKeys != null) {
-        if (generatedKeys.getResultSet() != null) {
-          generatedKeys.getResultSet().close();
+        ResultSet resultSet = generatedKeys.getResultSet();
+        if (resultSet != null) {
+          resultSet.close();
         }
-        generatedKeys = null;
+        this.generatedKeys = null;
       }
     }
   }
@@ -352,7 +382,7 @@ public class PgStatement implements Statement, BaseStatement {
    * @param cachedQuery to check (null if current query)
    * @return true if query is unlikely to be reused
    */
-  protected boolean isOneShotQuery(CachedQuery cachedQuery) {
+  protected boolean isOneShotQuery(@Nullable CachedQuery cachedQuery) {
     if (cachedQuery == null) {
       return true;
     }
@@ -364,7 +394,8 @@ public class PgStatement implements Statement, BaseStatement {
     return false;
   }
 
-  protected final void execute(CachedQuery cachedQuery, ParameterList queryParameters, int flags)
+  protected final void execute(CachedQuery cachedQuery,
+      @Nullable ParameterList queryParameters, int flags)
       throws SQLException {
     try {
       executeInternal(cachedQuery, queryParameters, flags);
@@ -380,7 +411,8 @@ public class PgStatement implements Statement, BaseStatement {
     }
   }
 
-  private void executeInternal(CachedQuery cachedQuery, ParameterList queryParameters, int flags)
+  private void executeInternal(CachedQuery cachedQuery,
+      @Nullable ParameterList queryParameters, int flags)
       throws SQLException {
     closeForNextExecution();
 
@@ -410,6 +442,9 @@ public class PgStatement implements Statement, BaseStatement {
     if (connection.getAutoCommit()) {
       flags |= QueryExecutor.QUERY_SUPPRESS_BEGIN;
     }
+    if (connection.hintReadOnly()) {
+      flags |= QueryExecutor.QUERY_READ_ONLY_HINT;
+    }
 
     // updateable result sets do not yet support binary updates
     if (concurrency != ResultSet.CONCUR_READ_ONLY) {
@@ -433,7 +468,7 @@ public class PgStatement implements Statement, BaseStatement {
           flags2);
       ResultWrapper result2 = handler2.getResults();
       if (result2 != null) {
-        result2.getResultSet().close();
+        castNonNull(result2.getResultSet(), "result2.getResultSet()").close();
       }
     }
 
@@ -450,11 +485,13 @@ public class PgStatement implements Statement, BaseStatement {
     }
     synchronized (this) {
       checkClosed();
-      result = firstUnclosedResult = handler.getResults();
+
+      ResultWrapper currentResult = handler.getResults();
+      result = firstUnclosedResult = currentResult;
 
       if (wantsGeneratedKeysOnce || wantsGeneratedKeysAlways) {
-        generatedKeys = result;
-        result = result.getNext();
+        generatedKeys = currentResult;
+        result = castNonNull(currentResult, "handler.getResults()").getNext();
 
         if (wantsGeneratedKeysOnce) {
           wantsGeneratedKeysOnce = false;
@@ -470,6 +507,7 @@ public class PgStatement implements Statement, BaseStatement {
 
   private volatile boolean isClosed = false;
 
+  @Override
   public int getUpdateCount() throws SQLException {
     synchronized (this) {
       checkClosed();
@@ -477,29 +515,13 @@ public class PgStatement implements Statement, BaseStatement {
         return -1;
       }
 
-      return result.getUpdateCount();
+      long count = result.getUpdateCount();
+      return count > Integer.MAX_VALUE ? Statement.SUCCESS_NO_INFO : (int) count;
     }
   }
 
   public boolean getMoreResults() throws SQLException {
-    synchronized (this) {
-      checkClosed();
-      if (result == null) {
-        return false;
-      }
-
-      result = result.getNext();
-
-      // Close preceding resultsets.
-      while (firstUnclosedResult != result) {
-        if (firstUnclosedResult.getResultSet() != null) {
-          firstUnclosedResult.getResultSet().close();
-        }
-        firstUnclosedResult = firstUnclosedResult.getNext();
-      }
-
-      return (result != null && result.getResultSet() != null);
-    }
+    return getMoreResults(CLOSE_ALL_RESULTS);
   }
 
   public int getMaxRows() throws SQLException {
@@ -582,7 +604,7 @@ public class PgStatement implements Statement, BaseStatement {
     }
   }
 
-  public SQLWarning getWarnings() throws SQLException {
+  public @Nullable SQLWarning getWarnings() throws SQLException {
     checkClosed();
     //copy reference to avoid NPE from concurrent modification of this.warnings
     final PSQLWarningWrapper warnWrap = this.warnings;
@@ -616,7 +638,7 @@ public class PgStatement implements Statement, BaseStatement {
     warnings = null;
   }
 
-  public ResultSet getResultSet() throws SQLException {
+  public @Nullable ResultSet getResultSet() throws SQLException {
     synchronized (this) {
       checkClosed();
 
@@ -713,9 +735,13 @@ public class PgStatement implements Statement, BaseStatement {
   public void addBatch(String sql) throws SQLException {
     checkClosed();
 
+    ArrayList<Query> batchStatements = this.batchStatements;
     if (batchStatements == null) {
-      batchStatements = new ArrayList<Query>();
-      batchParameters = new ArrayList<ParameterList>();
+      this.batchStatements = batchStatements = new ArrayList<Query>();
+    }
+    ArrayList<@Nullable ParameterList> batchParameters = this.batchParameters;
+    if (batchParameters == null) {
+      this.batchParameters = batchParameters = new ArrayList<@Nullable ParameterList>();
     }
 
     // Simple statements should not replace ?, ? with $1, $2
@@ -729,36 +755,32 @@ public class PgStatement implements Statement, BaseStatement {
   public void clearBatch() throws SQLException {
     if (batchStatements != null) {
       batchStatements.clear();
+    }
+    if (batchParameters != null) {
       batchParameters.clear();
     }
   }
 
   protected BatchResultHandler createBatchHandler(Query[] queries,
-      ParameterList[] parameterLists) {
+      @Nullable ParameterList[] parameterLists) {
     return new BatchResultHandler(this, queries, parameterLists,
         wantsGeneratedKeysAlways);
   }
 
-  public int[] executeBatch() throws SQLException {
-    checkClosed();
-
-    closeForNextExecution();
-
-    if (batchStatements == null || batchStatements.isEmpty()) {
-      return new int[0];
-    }
-
+  @RequiresNonNull({"batchStatements", "batchParameters"})
+  private BatchResultHandler internalExecuteBatch() throws SQLException {
     // Construct query/parameter arrays.
     transformQueriesAndParameters();
+    ArrayList<Query> batchStatements = castNonNull(this.batchStatements);
+    ArrayList<@Nullable ParameterList> batchParameters = castNonNull(this.batchParameters);
     // Empty arrays should be passed to toArray
     // see http://shipilev.net/blog/2016/arrays-wisdom-ancients/
     Query[] queries = batchStatements.toArray(new Query[0]);
-    ParameterList[] parameterLists =
-        batchParameters.toArray(new ParameterList[0]);
+    @Nullable ParameterList[] parameterLists = batchParameters.toArray(new ParameterList[0]);
     batchStatements.clear();
     batchParameters.clear();
 
-    int flags = 0;
+    int flags;
 
     // Force a Describe before any execution? We need to do this if we're going
     // to send anything dependent on the Describe results, e.g. binary parameters.
@@ -819,6 +841,9 @@ public class PgStatement implements Statement, BaseStatement {
     if (connection.getAutoCommit()) {
       flags |= QueryExecutor.QUERY_SUPPRESS_BEGIN;
     }
+    if (connection.hintReadOnly()) {
+      flags |= QueryExecutor.QUERY_READ_ONLY_HINT;
+    }
 
     BatchResultHandler handler;
     handler = createBatchHandler(queries, parameterLists);
@@ -840,7 +865,7 @@ public class PgStatement implements Statement, BaseStatement {
       }
       ResultWrapper result2 = handler2.getResults();
       if (result2 != null) {
-        result2.getResultSet().close();
+        castNonNull(result2.getResultSet(), "result2.getResultSet()").close();
       }
     }
 
@@ -862,8 +887,18 @@ public class PgStatement implements Statement, BaseStatement {
         }
       }
     }
+    return handler;
+  }
 
-    return handler.getUpdateCount();
+  public int[] executeBatch() throws SQLException {
+    checkClosed();
+    closeForNextExecution();
+
+    if (batchStatements == null || batchStatements.isEmpty() || batchParameters == null) {
+      return new int[0];
+    }
+
+    return internalExecuteBatch().getUpdateCount();
   }
 
   public void cancel() throws SQLException {
@@ -915,7 +950,7 @@ public class PgStatement implements Statement, BaseStatement {
     }
   }
 
-  public void setFetchSize(int rows) throws SQLException {
+  public void setFetchSize(@NonNegative int rows) throws SQLException {
     checkClosed();
     if (rows < 0) {
       throw new PSQLException(GT.tr("Fetch size must be a value greater to or equal to 0."),
@@ -1012,8 +1047,17 @@ public class PgStatement implements Statement, BaseStatement {
     return forceBinaryTransfers;
   }
 
+  //#if mvn.project.property.postgresql.jdbc.spec >= "JDBC4.2"
+  @Override
   public long getLargeUpdateCount() throws SQLException {
-    throw Driver.notImplemented(this.getClass(), "getLargeUpdateCount");
+    synchronized (this) {
+      checkClosed();
+      if (result == null || result.getResultSet() != null) {
+        return -1;
+      }
+
+      return result.getUpdateCount();
+    }
   }
 
   public void setLargeMaxRows(long max) throws SQLException {
@@ -1024,29 +1068,57 @@ public class PgStatement implements Statement, BaseStatement {
     throw Driver.notImplemented(this.getClass(), "getLargeMaxRows");
   }
 
+  @Override
   public long[] executeLargeBatch() throws SQLException {
-    throw Driver.notImplemented(this.getClass(), "executeLargeBatch");
+    checkClosed();
+    closeForNextExecution();
+
+    if (batchStatements == null || batchStatements.isEmpty() || batchParameters == null) {
+      return new long[0];
+    }
+
+    return internalExecuteBatch().getLargeUpdateCount();
   }
 
+  @Override
   public long executeLargeUpdate(String sql) throws SQLException {
-    throw Driver.notImplemented(this.getClass(), "executeLargeUpdate");
+    executeWithFlags(sql, QueryExecutor.QUERY_NO_RESULTS);
+    checkNoResultUpdate();
+    return getLargeUpdateCount();
   }
 
+  @Override
   public long executeLargeUpdate(String sql, int autoGeneratedKeys) throws SQLException {
-    throw Driver.notImplemented(this.getClass(), "executeLargeUpdate");
+    if (autoGeneratedKeys == Statement.NO_GENERATED_KEYS) {
+      return executeLargeUpdate(sql);
+    }
+
+    return executeLargeUpdate(sql, (String[]) null);
   }
 
+  @Override
   public long executeLargeUpdate(String sql, int[] columnIndexes) throws SQLException {
-    throw Driver.notImplemented(this.getClass(), "executeLargeUpdate");
+    if (columnIndexes == null || columnIndexes.length == 0) {
+      return executeLargeUpdate(sql);
+    }
+
+    throw new PSQLException(GT.tr("Returning autogenerated keys by column index is not supported."),
+        PSQLState.NOT_IMPLEMENTED);
   }
 
-  public long executeLargeUpdate(String sql, String[] columnNames) throws SQLException {
-    throw Driver.notImplemented(this.getClass(), "executeLargeUpdate");
-  }
+  @Override
+  public long executeLargeUpdate(String sql, String @Nullable [] columnNames) throws SQLException {
+    if (columnNames != null && columnNames.length == 0) {
+      return executeLargeUpdate(sql);
+    }
 
-  public long executeLargeUpdate() throws SQLException {
-    throw Driver.notImplemented(this.getClass(), "executeLargeUpdate");
+    wantsGeneratedKeysOnce = true;
+    if (!executeCachedSql(sql, 0, columnNames)) {
+      // no resultset returned. What's a pity!
+    }
+    return getLargeUpdateCount();
   }
+  //#endif
 
   public boolean isClosed() throws SQLException {
     return isClosed;
@@ -1089,7 +1161,8 @@ public class PgStatement implements Statement, BaseStatement {
     synchronized (this) {
       ResultWrapper result = firstUnclosedResult;
       while (result != null) {
-        if (result.getResultSet() != null && !result.getResultSet().isClosed()) {
+        ResultSet resultSet = result.getResultSet();
+        if (resultSet != null && !resultSet.isClosed()) {
           return;
         }
         result = result.getNext();
@@ -1123,12 +1196,7 @@ public class PgStatement implements Statement, BaseStatement {
       // CLOSE_ALL_RESULTS
       if (current == Statement.CLOSE_ALL_RESULTS) {
         // Close preceding resultsets.
-        while (firstUnclosedResult != result) {
-          if (firstUnclosedResult.getResultSet() != null) {
-            firstUnclosedResult.getResultSet().close();
-          }
-          firstUnclosedResult = firstUnclosedResult.getNext();
-        }
+        closeUnclosedProcessedResults();
       }
 
       // Done.
@@ -1140,7 +1208,7 @@ public class PgStatement implements Statement, BaseStatement {
     synchronized (this) {
       checkClosed();
       if (generatedKeys == null || generatedKeys.getResultSet() == null) {
-        return createDriverResultSet(new Field[0], new ArrayList<byte[][]>());
+        return createDriverResultSet(new Field[0], new ArrayList<Tuple>());
       }
 
       return generatedKeys.getResultSet();
@@ -1164,7 +1232,7 @@ public class PgStatement implements Statement, BaseStatement {
         PSQLState.NOT_IMPLEMENTED);
   }
 
-  public int executeUpdate(String sql, String[] columnNames) throws SQLException {
+  public int executeUpdate(String sql, String @Nullable [] columnNames) throws SQLException {
     if (columnNames != null && columnNames.length == 0) {
       return executeUpdate(sql);
     }
@@ -1183,7 +1251,7 @@ public class PgStatement implements Statement, BaseStatement {
     return execute(sql, (String[]) null);
   }
 
-  public boolean execute(String sql, int[] columnIndexes) throws SQLException {
+  public boolean execute(String sql, int @Nullable [] columnIndexes) throws SQLException {
     if (columnIndexes != null && columnIndexes.length == 0) {
       return execute(sql);
     }
@@ -1192,7 +1260,7 @@ public class PgStatement implements Statement, BaseStatement {
         PSQLState.NOT_IMPLEMENTED);
   }
 
-  public boolean execute(String sql, String[] columnNames) throws SQLException {
+  public boolean execute(String sql, String @Nullable [] columnNames) throws SQLException {
     if (columnNames != null && columnNames.length == 0) {
       return execute(sql);
     }
@@ -1205,7 +1273,7 @@ public class PgStatement implements Statement, BaseStatement {
     return rsHoldability;
   }
 
-  public ResultSet createDriverResultSet(Field[] fields, List<byte[][]> tuples)
+  public ResultSet createDriverResultSet(Field[] fields, List<Tuple> tuples)
       throws SQLException {
     return createResultSet(null, fields, tuples, null);
   }
