@@ -6,12 +6,18 @@
 package org.postgresql.test;
 
 import org.postgresql.PGProperty;
+import org.postgresql.core.BaseConnection;
 import org.postgresql.core.ServerVersion;
+import org.postgresql.core.TransactionState;
 import org.postgresql.core.Version;
+import org.postgresql.jdbc.GSSEncMode;
 import org.postgresql.jdbc.PgConnection;
+import org.postgresql.util.PSQLException;
 
+import org.checkerframework.checker.nullness.qual.Nullable;
 import org.junit.Assert;
 
+import java.io.Closeable;
 import java.io.File;
 import java.io.FileInputStream;
 import java.io.IOException;
@@ -165,6 +171,13 @@ public class TestUtil {
     return System.getProperty(PGProperty.SSL_PASSWORD.getName());
   }
 
+  /*
+   *  Return the GSSEncMode for the tests
+   */
+  public static GSSEncMode getGSSEncMode() throws PSQLException {
+    return GSSEncMode.of(System.getProperties());
+  }
+
   /*
    * Returns the user for SSPI authentication tests
    */
@@ -294,6 +307,8 @@ public class TestUtil {
   public static Connection openPrivilegedDB() throws SQLException {
     initDriver();
     Properties properties = new Properties();
+
+    PGProperty.GSS_ENC_MODE.set(properties,getGSSEncMode().value);
     properties.setProperty("user", getPrivilegedUser());
     properties.setProperty("password", getPrivilegedPassword());
     return DriverManager.getConnection(getURL(), properties);
@@ -326,11 +341,14 @@ public class TestUtil {
           "user name is not specified. Please specify 'username' property via -D or build.properties");
     }
     props.setProperty("user", user);
-    String password = getPassword();
+
+    // Allow properties to override the password.
+    String password = props.getProperty("password");
     if (password == null) {
-      password = "";
+      password = getPassword() != null ? getPassword() : "";
     }
     props.setProperty("password", password);
+
     String sslPassword = getSslPassword();
     if (sslPassword != null) {
       PGProperty.SSL_PASSWORD.set(props, sslPassword);
@@ -349,13 +367,18 @@ public class TestUtil {
     String hostport = props.getProperty(SERVER_HOST_PORT_PROP, getServer() + ":" + getPort());
     String database = props.getProperty(DATABASE_PROP, getDatabase());
 
+    // Set GSSEncMode for tests only in the case the property is already missing
+    if (PGProperty.GSS_ENC_MODE.getSetString(props) == null) {
+      PGProperty.GSS_ENC_MODE.set(props, getGSSEncMode().value);
+    }
+
     return DriverManager.getConnection(getURL(hostport, database), props);
   }
 
   /*
    * Helper - closes an open connection.
    */
-  public static void closeDB(Connection con) throws SQLException {
+  public static void closeDB(@Nullable Connection con) throws SQLException {
     if (con != null) {
       con.close();
     }
@@ -383,18 +406,7 @@ public class TestUtil {
    * Helper - drops a schema
    */
   public static void dropSchema(Connection con, String schema) throws SQLException {
-    Statement stmt = con.createStatement();
-    try {
-      if (con.getAutoCommit()) {
-        // Not in a transaction so ignore error for missing object
-        stmt.executeUpdate("DROP SCHEMA IF EXISTS " + schema + " CASCADE");
-      } else {
-        // In a transaction so do not ignore errors for missing object
-        stmt.executeUpdate("DROP SCHEMA " + schema + " CASCADE");
-      }
-    } finally {
-      closeQuietly(stmt);
-    }
+    dropObject(con, "SCHEMA", schema);
   }
 
   /*
@@ -422,7 +434,6 @@ public class TestUtil {
    * @param table String
    * @param columns String
    */
-
   public static void createTempTable(Connection con, String table, String columns)
       throws SQLException {
     Statement st = con.createStatement();
@@ -437,6 +448,44 @@ public class TestUtil {
     }
   }
 
+  /*
+   * Helper - creates a unlogged table for use by a test.
+   * Unlogged tables works from PostgreSQL 9.1+
+   */
+  public static void createUnloggedTable(Connection con, String table, String columns)
+      throws SQLException {
+    Statement st = con.createStatement();
+    try {
+      // Drop the table
+      dropTable(con, table);
+
+      String unlogged = haveMinimumServerVersion(con, ServerVersion.v9_1) ? "UNLOGGED" : "";
+
+      // Now create the table
+      st.executeUpdate("CREATE " + unlogged + " TABLE " + table + " (" + columns + ")");
+    } finally {
+      closeQuietly(st);
+    }
+  }
+
+  /*
+   * Helper - creates a view
+   */
+  public static void createView(Connection con, String viewName, String query)
+      throws SQLException {
+    Statement st = con.createStatement();
+    try {
+      // Drop the view
+      dropView(con, viewName);
+
+      String sql = "CREATE VIEW " + viewName + " AS " + query;
+
+      st.executeUpdate(sql);
+    } finally {
+      closeQuietly(st);
+    }
+  }
+
   /**
    * Helper creates an enum type.
    *
@@ -444,14 +493,12 @@ public class TestUtil {
    * @param name String
    * @param values String
    */
-
   public static void createEnumType(Connection con, String name, String values)
       throws SQLException {
     Statement st = con.createStatement();
     try {
       dropType(con, name);
 
-
       // Now create the table
       st.executeUpdate("create type " + name + " as enum (" + values + ")");
     } finally {
@@ -495,22 +542,11 @@ public class TestUtil {
    * Drops a domain.
    *
    * @param con Connection
-   * @param name String
+   * @param domain String
    */
-  public static void dropDomain(Connection con, String name)
+  public static void dropDomain(Connection con, String domain)
       throws SQLException {
-    Statement stmt = con.createStatement();
-    try {
-      if (con.getAutoCommit()) {
-        // Not in a transaction so ignore error for missing object
-        stmt.executeUpdate("DROP DOMAIN IF EXISTS " + name + " CASCADE");
-      } else {
-        // In a transaction so do not ignore errors for missing object
-        stmt.executeUpdate("DROP DOMAIN " + name + " CASCADE");
-      }
-    } finally {
-      closeQuietly(stmt);
-    }
+    dropObject(con, "DOMAIN", domain);
   }
 
   /**
@@ -536,50 +572,39 @@ public class TestUtil {
    * drop a sequence because older versions don't have dependency information for serials
    */
   public static void dropSequence(Connection con, String sequence) throws SQLException {
-    Statement stmt = con.createStatement();
-    try {
-      if (con.getAutoCommit()) {
-        // Not in a transaction so ignore error for missing object
-        stmt.executeUpdate("DROP SEQUENCE IF EXISTS " + sequence + " CASCADE");
-      } else {
-        // In a transaction so do not ignore errors for missing object
-        stmt.executeUpdate("DROP SEQUENCE " + sequence + " CASCADE");
-      }
-    } finally {
-      closeQuietly(stmt);
-    }
+    dropObject(con, "SEQUENCE", sequence);
   }
 
   /*
    * Helper - drops a table
    */
   public static void dropTable(Connection con, String table) throws SQLException {
-    Statement stmt = con.createStatement();
-    try {
-      if (con.getAutoCommit()) {
-        // Not in a transaction so ignore error for missing object
-        stmt.executeUpdate("DROP TABLE IF EXISTS " + table + " CASCADE ");
-      } else {
-        // In a transaction so do not ignore errors for missing object
-        stmt.executeUpdate("DROP TABLE " + table + " CASCADE ");
-      }
-    } finally {
-      closeQuietly(stmt);
-    }
+    dropObject(con, "TABLE", table);
+  }
+
+  /*
+   * Helper - drops a view
+   */
+  public static void dropView(Connection con, String view) throws SQLException {
+    dropObject(con, "VIEW", view);
   }
 
   /*
    * Helper - drops a type
    */
   public static void dropType(Connection con, String type) throws SQLException {
+    dropObject(con, "TYPE", type);
+  }
+
+  private static void dropObject(Connection con, String type, String name) throws SQLException {
     Statement stmt = con.createStatement();
     try {
       if (con.getAutoCommit()) {
         // Not in a transaction so ignore error for missing object
-        stmt.executeUpdate("DROP TYPE IF EXISTS " + type + " CASCADE");
+        stmt.executeUpdate("DROP " + type + " IF EXISTS " + name + " CASCADE");
       } else {
         // In a transaction so do not ignore errors for missing object
-        stmt.executeUpdate("DROP TYPE " + type + " CASCADE");
+        stmt.executeUpdate("DROP " + type + " " + name + " CASCADE");
       }
     } finally {
       closeQuietly(stmt);
@@ -601,6 +626,11 @@ public class TestUtil {
     }
   }
 
+  public static void assertTransactionState(String message, Connection con, TransactionState expected) {
+    TransactionState actual = TestUtil.getTransactionState(con);
+    Assert.assertEquals(message, expected, actual);
+  }
+
   /*
    * Helper - generates INSERT SQL - very simple
    */
@@ -784,10 +814,22 @@ public class TestUtil {
     return returnValue;
   }
 
+  /**
+   * Close a resource and ignore any errors during closing.
+   */
+  public static void closeQuietly(@Nullable Closeable resource) {
+    if (resource != null) {
+      try {
+        resource.close();
+      } catch (Exception ignore) {
+      }
+    }
+  }
+
   /**
    * Close a Connection and ignore any errors during closing.
    */
-  public static void closeQuietly(Connection conn) {
+  public static void closeQuietly(@Nullable Connection conn) {
     if (conn != null) {
       try {
         conn.close();
@@ -799,7 +841,7 @@ public class TestUtil {
   /**
    * Close a Statement and ignore any errors during closing.
    */
-  public static void closeQuietly(Statement stmt) {
+  public static void closeQuietly(@Nullable Statement stmt) {
     if (stmt != null) {
       try {
         stmt.close();
@@ -811,7 +853,7 @@ public class TestUtil {
   /**
    * Close a ResultSet and ignore any errors during closing.
    */
-  public static void closeQuietly(ResultSet rs) {
+  public static void closeQuietly(@Nullable ResultSet rs) {
     if (rs != null) {
       try {
         rs.close();
@@ -897,6 +939,73 @@ public class TestUtil {
     }
   }
 
+  /**
+   * Execute a SQL query with a given connection and return whether any rows were
+   * returned. No column data is fetched.
+   */
+  public static boolean executeQuery(Connection conn, String sql) throws SQLException {
+    Statement stmt = conn.createStatement();
+    ResultSet rs = stmt.executeQuery(sql);
+    boolean hasNext = rs.next();
+    rs.close();
+    stmt.close();
+    return hasNext;
+  }
+
+  /**
+   * Retrieve the backend process id for a given connection.
+   */
+  public static int getBackendPid(Connection conn) throws SQLException {
+    Statement stmt = conn.createStatement();
+    ResultSet rs = stmt.executeQuery("SELECT pg_backend_pid()");
+    rs.next();
+    int pid = rs.getInt(1);
+    rs.close();
+    stmt.close();
+    return pid;
+  }
+
+  /**
+   * Executed pg_terminate_backend(...) to terminate the server process for
+   * a given process id with the given connection.
+   */
+  public static boolean terminateBackend(Connection conn, int backendPid) throws SQLException {
+    PreparedStatement stmt = conn.prepareStatement("SELECT pg_terminate_backend(?)");
+    stmt.setInt(1, backendPid);
+    ResultSet rs = stmt.executeQuery();
+    rs.next();
+    boolean wasTerminated = rs.getBoolean(1);
+    rs.close();
+    stmt.close();
+    return wasTerminated;
+  }
+
+  /**
+   * Create a new connection using the default test credentials and use it to
+   * attempt to terminate the specified backend process.
+   */
+  public static boolean terminateBackend(int backendPid) throws SQLException {
+    Connection conn = TestUtil.openPrivilegedDB();
+    try {
+      return terminateBackend(conn, backendPid);
+    } finally {
+      conn.close();
+    }
+  }
+
+  /**
+   * Retrieve the given connection backend process id, then create a new connection
+   * using the default test credentials and attempt to terminate the process.
+   */
+  public static boolean terminateBackend(Connection conn) throws SQLException {
+    int pid = getBackendPid(conn);
+    return terminateBackend(pid);
+  }
+
+  public static TransactionState getTransactionState(Connection conn) {
+    return ((BaseConnection) conn).getTransactionState();
+  }
+
   private static void waitStopReplicationSlot(Connection connection, String slotName)
       throws InterruptedException, TimeoutException, SQLException {
     long startWaitTime = System.currentTimeMillis();
