@@ -19,6 +19,8 @@ import org.junit.runners.Suite;
     BlobTest.class,
     CharacterStreamTest.class,
     ClientInfoTest.class,
+    ConnectionValidTimeoutTest.class,
+    DatabaseMetaDataHideUnprivilegedObjectsTest.class,
     DatabaseMetaDataTest.class,
     IsValidTest.class,
     JsonbTest.class,
@@ -30,4 +32,3 @@ import org.junit.runners.Suite;
 })
 public class Jdbc4TestSuite {
 }
-
